#define BOOST_BIND_GLOBAL_PLACEHOLDERS 1
#ifdef WIN32
#define _SILENCE_CXX17_CODECVT_HEADER_DEPRECATION_WARNING
#define _SILENCE_ALL_CXX17_DEPRECATION_WARNINGS
#endif

#ifndef WIN32
#include <csignal>
#include <unistd.h>
#endif

#include <iostream>
#include <string>
#include <cstring>
#include <stdexcept>
#include <fstream>
#include <chrono>
#include <thread>
#include <streambuf>
#include <sstream>
#include <csignal>
#include <filesystem>

#include <boost/bind/bind.hpp>
#include <boost/property_tree/ptree.hpp>
#include <boost/property_tree/json_parser.hpp>

#include <fcf/format/json/json.hpp>
#include <fcf/log.hpp>
#include <fcf/net/net.hpp>

#include "configuration.hpp"
#include "server.hpp"

application::server_sptr_t pserver_g;
  
void run(std::filesystem::path a_config_path) {
  application::server_configuration_t sc;

  fcf::net::initialize();

  std::ifstream file(a_config_path);
  std::stringstream ss(fcf::format::json::adapt_json(std::string((std::istreambuf_iterator<char>(file)), std::istreambuf_iterator<char>())));
  if (ss.str().empty())
    throw std::runtime_error("Failed to read config file.");
  try {
    boost::property_tree::ptree conf_tree;
    boost::property_tree::read_json(ss, conf_tree);
    sc.set_values(conf_tree);
  } catch(const std::exception& e){
    throw std::runtime_error(std::string("Incorrect config file. ") + e.what());
  }

  fcf::log.get_logger().set_level(sc.logLevel);
  fcf::log.get_logger().set_file_prefix(sc.logFile);
  fcf::log.get_logger().set_life_time(sc.logLifeTime);

  fcf::log << "--------------------------------------------";
  fcf::log << "Start server";

  pserver_g.reset(new application::server_t(sc));
  pserver_g->run();
}


#ifdef WIN32
  void int_signal_handler(int signum) {
    if (pserver_g.get())
      pserver_g->stop();
    exit(0);
  }

  int wmain(int a_argc, const wchar_t* a_argv[]){
    try {
      if (a_argc < 2)
        throw std::runtime_error("Config file not specified for command line parameter");
      run(std::filesystem::path(a_argv[1]));
      signal(SIGINT, int_signal_handler);
      while (true)
        std::this_thread::sleep_for(std::chrono::seconds(1));
    }
    catch (const std::exception& e) {
      fcf::err << "[Failed start] " << e.what();
      try {
        if (pserver_g.get())
          pserver_g->stop();      
      } catch(...) {
      }
      return 1;
    }
    return 0;
  }

#else

  #include <sys/types.h>
  #include <sys/wait.h>


  void int_signal_handler(int) {
    if (pserver_g.get())
      pserver_g->stop();
    exit(0);
  }

  void int_signal_child_handler(int) {
    int status;
    pid_t pid;
    pid = wait(&status);
  }

  void broken_pipe_signal_handler(int){
  }

  int main(int a_argc, const char* a_argv[]) {
    signal(SIGPIPE, broken_pipe_signal_handler);
    signal(SIGCHLD, int_signal_child_handler);
    try {
      if (a_argc < 2)
        throw std::runtime_error("Config file not specified for command line parameter");    
      run(std::filesystem::path(a_argv[1]));
      signal(SIGINT, int_signal_handler);
      
      
      while(true)
        std::this_thread::sleep_for(std::chrono::seconds(1));
    } catch (const std::exception& e) {
      fcf::err << "[Failed start] " << e.what();
      try {
        if (pserver_g.get())
          pserver_g->stop();      
      } catch(...) {
      }
      return 1;
    }

    return 0;
  }

#endif
